// pari_mpl.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"

BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
					 )
{
    return TRUE;
}

#include "stdafx.h"
#include <malloc.h>
#include <memory.h>
#include <string.h>

#include "pari.h"
#include "paripriv.h"
#include "parigen.h"

long PARIinitialized;

#define maximalLength 100000 // length of strings for PARI in or out, approx 20 pages
//long resultNumeric;
//long precisionPARI;


/*******************************************************************
	settings: Open Maple
*******************************************************************/

#include <stdlib.h>
#include <string.h>

#include <maplec.h>
#include <mplshlib.h>
#include <mpltable.h>

#if !defined(EXPORT_DECL)
#  ifdef _MSC_VER
#    define EXPORT_DECL __declspec(dllexport)
#  else
#    define EXPORT_DECL
#  endif
#endif

/*
	eval_pari:
	Take a command as string for the first argument, hand it over to PARI to evaluate
	it. Write the result (converted to a string) in the second argument, if no errors
	occured and return the length of the resulting string.
	In case of numerical results there is some string manipulation, s.th. Maple finally
	should be able to parse that to Maple numerics. The global resultNumeric is set to 1.
	In case of an error write PARI's error message to the second argument and return
	the _negative_ length of that string. The global resultNumeric is set to 0.

	Assumes that Pari is initialized and resultStr has (enough) allocated memory
*/
long eval_pari(char* resultStr, char* inputStr)
{
	GEN y;
	char* outputBuf;
	long inputLen = 0;
	long outputLen = 0;
	char* p_error = " [PARI error message]:\n"; 

	//resultNumeric = 0;

	inputLen = strlen(inputStr);
	if ( inputLen <= 0 )
		{
		memmove(resultStr, "\0", 1);	// resultStr[0] =  "\0";
		return 0;
		}

	outputBuf = (char *)calloc( maximalLength, sizeof( char ) );;

	CATCH(CATCH_ALL)
  { /* we have caught an exception ! */
    CATCH_RELEASE();
		free(outputBuf);
		//memmove(resultStr, "\0", 1);
		outputLen = strlen(errmessage[pari_errno]);
		memmove(resultStr, p_error, strlen(p_error));
		strcat(resultStr, errmessage[pari_errno]); 
		//memmove(resultStr, errmessage[pari_errno], outputLen);
		return - (outputLen + strlen(p_error)); // negative, pari error message
    //... do something to abort gracefully. A priori a longjmp ...
    //'pari_errno' contains the error number.
  }
	TRY
  {
	y = gp_read_str(inputStr);	// compute with PARI
	//y = gtofp(y, precision(dbltor(0.0)));		// floating point evaluation
	brute(y,'e',-1);						// raw format: libpari.pdf, 4.7.2 output, p.29 & users.pdf, 2.11.16 output, p.58
	outputBuf = GENtostr( y );	// revert to C string

	outputLen = strlen( outputBuf );
  }
  ENDCATCH;

	if ( outputLen <= 0 )
		{
		memmove(resultStr, inputStr, inputLen);
		free(outputBuf);
		return 0;
		}

	memmove(resultStr, outputBuf, outputLen);
	free(outputBuf);
	return outputLen;
}

EXPORT_DECL ALGEB M_DECL start_Pari( MKernelVector kv,	ALGEB fn_args )
{
	if (PARIinitialized == 1){;}	// already alive? else start or restart? hm ...
	else
	{
		pari_init(1000000,2);
		MaplePrintf(kv,"PARI 2.3.0 initialized.");
		PARIinitialized = 1;
	}
	return( ToMapleInteger(kv,PARIinitialized) );
}

EXPORT_DECL ALGEB M_DECL stop_Pari( MKernelVector kv,	ALGEB fn_args )
{
	if (PARIinitialized == 0){;}	// already already killed? else shut down
	else
	{
		pari_close();
		MaplePrintf(kv,"PARI 2.3.0 closed.");
		PARIinitialized = 0;
	}
	return( ToMapleInteger(kv,PARIinitialized) );
}

/*
	Main entry for the DLL to be called from Maple

	usage stuff:=define_external(`eval_Pari`,MAPLE,LIB=theDLL); and then
	stuff(somePARIstring), where somePARIstring is a string in Maple which
	PARI should to evaluate (like stuff("exp(1)") ).

	That uses OpenMaple and the function eval_pari to convert strings between
	Maple and C formats.

	For compiling one needs a (complete) Maple installation.
*/

EXPORT_DECL ALGEB M_DECL eval_Pari( MKernelVector kv,	ALGEB fn_args )
{
	ALGEB *args;
	ALGEB someProc;
	ALGEB aStrMaple;
	ALGEB result;

	long length_Maple;

	char* inBuf;
	long inLen;
	char* outBuf;
	long outLen;

  args = (ALGEB*) fn_args;

	// simple input validation
	if( MapleNumArgs(kv,(ALGEB)args) != 1 )
		{
		MapleRaiseError(kv,"expecting 1 argument: a string");
		return( NULL );
		}

	aStrMaple = args[1];
	if( !IsMapleString(kv, aStrMaple) )
		{
		MapleRaiseError(kv,"expecting 1 argument: a string");
		return( NULL );
		}

	// convert a Maple string to a C string,
	// care for length to avoid over-writing memory

	someProc = EvalMapleStatement(kv,"length:");
	length_Maple = MapleToInteger32(kv, EvalMapleProc(kv,someProc,1, aStrMaple));

	inBuf = (char *)calloc( maximalLength, sizeof( char ) );
	inLen = min(maximalLength,(long)length_Maple);
	strncpy(inBuf, MapleToString(kv,args[1]), inLen);

	outBuf = (char *)calloc( maximalLength, sizeof( char ) );

	if (PARIinitialized == 1){;}	// already alive? else start with defaults
	else
	{
		pari_init(1000000,2);
		PARIinitialized = 1;
	}

	outLen = eval_pari(outBuf,inBuf);	// calling Pari to evaluate
	free(inBuf);

	//pari_close();

	if( outLen < 0 )
		{
		outBuf = realloc( outBuf, (1000 * sizeof( char )) );
		//strcat(outBuf, "\n[PARI error message]");
		result = ToMapleString(kv, outBuf);
		MapleRaiseError(kv,outBuf);
		free(outBuf);
		return( NULL );
		}

	if( outLen <= 0 )
		result = ToMapleString(kv,"");	//MapleRaiseError(KV,"some shit happend");

	result = ToMapleString(kv,outBuf);
	free(outBuf);

	return ( result );
}